package cz.muni.pa165.service;

import cz.muni.pa165.data.enums.ApplicationStatusEnum;
import cz.muni.pa165.data.model.Application;
import cz.muni.pa165.data.repository.ApplicationRepository;
import cz.muni.pa165.exceptions.ResourceNotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.time.LocalDate;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * @author Michal Badin
 */
@Service
public class ApplicationService {
    private static final Logger log = LoggerFactory.getLogger(ApplicationService.class);
    private static final String NOTIFICATION_MODULE_URL_NEW = "http://localhost:8083/notification/application/new";
    private static final String NOTIFICATION_MODULE_URL_STATUS = "http://localhost:8083/notification/application/status";
    private static final List<String> NOTIFICATION_RECEIVERS = List.of("formula.team.management@gmail.com");

    private final ApplicationRepository applicationRepository;
    private final RestTemplate restTemplate;

    @Autowired
    public ApplicationService(ApplicationRepository applicationRepository, RestTemplate restTemplate) {
        this.applicationRepository = applicationRepository;
        this.restTemplate = restTemplate;
    }

    @Transactional(readOnly = true)
    public List<Application> findAll() {
        return applicationRepository.findAll();
    }

    public Application create(Application application) {
        application.setFillingOutDate(LocalDate.now());
        application.setStatus(ApplicationStatusEnum.PENDING);

        var savedApplication = applicationRepository.save(application);

        try {
            sendPostRequest(savedApplication, NOTIFICATION_MODULE_URL_NEW, List.of(savedApplication.getEmail()));
        } catch (RestClientException | IllegalArgumentException e) {
            log.debug(String.format("The notification module is not reachable on the URL: %s", NOTIFICATION_MODULE_URL_NEW));
        }

        return savedApplication;
    }

    @Transactional(readOnly = true)
    public Application findById(Long id) {
        return applicationRepository.findById(id)
                .orElseThrow(() ->
                        new ResourceNotFoundException(String.format("%s with id %s was not found", Application.class, id)));
    }

    public Application setStatus(Long id, ApplicationStatusEnum applicationStatusEnum) {
        Optional<Application> applicationFromDb = applicationRepository.findById(id);
        if (applicationFromDb.isEmpty()) {
            throw new ResourceNotFoundException(String.format("%s with id %s was not found", Application.class, id));
        }

        applicationFromDb.get().setStatus(applicationStatusEnum);

        var savedApplication = applicationRepository.save(applicationFromDb.get());

        if (savedApplication.getStatus() != ApplicationStatusEnum.PENDING) {
            try {
                sendPostRequest(savedApplication, NOTIFICATION_MODULE_URL_STATUS, NOTIFICATION_RECEIVERS);
            } catch (RestClientException | IllegalArgumentException e) {
                log.debug(String.format("The notification module is not reachable on the URL: %s", NOTIFICATION_MODULE_URL_NEW));
            }
        }

        return savedApplication;
    }

    private void sendPostRequest(Application application, String url, List<String> receivers) {
        Map<String, Object> notification = new HashMap<>();
        notification.put("application", application);
        notification.put("receivers", receivers);

        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<Map<String, Object>> request = new HttpEntity<>(notification, headers);
        restTemplate.postForObject(url, request, String.class);
    }
}
