package cz.muni.pa165.service;

import cz.muni.pa165.data.model.Driver;
import cz.muni.pa165.data.repository.DriverRepository;
import cz.muni.pa165.exceptions.BadRequestException;
import cz.muni.pa165.exceptions.ResourceNotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * @author Michal Badin
 */
@Service
public class DriverService extends DomainService<Driver> {

    private static final Logger log = LoggerFactory.getLogger(DriverService.class);

    private final DriverRepository driverRepository;

    private final RestTemplate restTemplate;

    private static final String VISUALIZATION_MODULE_URL = "http://localhost:8082/visualization";

    @Autowired
    public DriverService(DriverRepository repository, RestTemplate restTemplate) {
        super(repository, Driver.class);
        driverRepository = repository;
        this.restTemplate = restTemplate;
    }

    @Transactional(readOnly = true)
    public Driver findById(Long id) {
        Optional<Driver> dbDriver = driverRepository.findById(id);
        if (dbDriver.isEmpty()) {
            throw new ResourceNotFoundException(entityClass, id);
        }

        try {
            sendPostRequest(dbDriver.get());
        } catch (RestClientException | IllegalArgumentException e) {
            log.debug(String.format("The visualization module is not reachable on the URL: %s", VISUALIZATION_MODULE_URL));
        }

        return dbDriver.get();
    }

    public Driver update(Long id, Driver driver) {
        Optional<Driver> dbDriver = driverRepository.findById(id);
        if (dbDriver.isEmpty()) {
            throw new ResourceNotFoundException(entityClass, id);
        }

        dbDriver.get().setName(driver.getName());
        dbDriver.get().setSurname(driver.getSurname());
        dbDriver.get().setNationality(driver.getNationality());
        dbDriver.get().setHeight(driver.getHeight());
        dbDriver.get().setCharacteristics(driver.getCharacteristics());
        return driverRepository.save(dbDriver.get());
    }

    @Transactional(readOnly = true)
    public List<Driver> findAllTestDrivers() {
        return driverRepository.findAllTestDrivers();
    }

    @Override
    public void delete(Long id) {
        var driverFromDb = repository.findById(id);
        if (driverFromDb.isEmpty()) {
            throw new ResourceNotFoundException(entityClass, id);
        }

        if (driverFromDb.get().getCar() != null) {
            throw new BadRequestException(
                    String.format("Driver with id %s is driver of the car with id %s. Remove the driver from the car first.",
                            id, driverFromDb.get().getCar().getId()));
        }

        repository.delete(driverFromDb.get());
    }

    private void sendPostRequest(Driver driver) {
        Map<String, Object> body = new HashMap<>();
        body.put("data", driver.toString());

        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<Map<String, Object>> request = new HttpEntity<>(body, headers);
        restTemplate.postForObject(VISUALIZATION_MODULE_URL, request, String.class);

        log.debug(String.format("Sent request %s to %s.", request, VISUALIZATION_MODULE_URL));
    }
}

